#include <p.h>

/* Path Integral Monte Carlo code developed by A. Balaz 
   (antun@ipb.ac.rs) for papers:
         
   "Systematic Speedup in the Convergence of Path Integrals"
   by A. Bogojevic, A. Balaz, and A. Belic
   Phys. Rev. Lett. 94 (2005) 180403

   "Fast Converging Path Integrals for Time-Dependent Potentials"
   by A. Balaz, I. Vidanovic, A. Bogojevic, A, Pelster
   arXiv:0912.2743

   Address: Institute of Physics, Belgrade, Serbia
   Scientific Computing Laboratory, http://www.scl.rs/speedup/

   Public use and modification of this code is allowed providing 
   the above paper is properly acknowledged. The author would be 
   grateful for all information and/or comments regarding the use   
   of the code.  
*/

/* USER CONFIGURABLE SECTION:

   1) Defining the level parameter p (1 <= p <= 8). 
   2) Supplying the C function for the evaluation of the potential V
      and its parameters.
*/

/* 1) Defining the level parameter p (1 <= p <= 8)

      The level parameter p corresponds to the function Vp()
      (improved potential) which will be compiled. You need to adjust
      the following preprocessor variable. The value below is for p = 4,
*/

#define _p 4

/* 2) Supplying the C function for the evaluation of the potential V
      and its parameters.

      Function V0(x) provides the value of the potential parameters,
      used later to calculate the effective potential.

      Parameters of the potential are par[0] ..., as entered by
      you on the command line.
*/

void V0(double x, double t) /* Example of the user supplied function V0() for the forced harmonic oscillator. */
{
   ss[1] = sin(par[0] * t);
   ss[2] = ss[1] * ss[1];
   cs[1] = cos(par[0] * t);
   cs[2] = cs[1] * cs[2];
   xs[1] = x;
   xs[2] = x * x;

   return;
}

/* END OF THE USER CONFIGURABLE SECTION! */


/* Function func(q[]) returns exp(- action + norm), where the variable action is
   the action for the given path (q[0], q[step], ..., q[N]) plus the exponent of
   the Gaussian probability density function (*distrexp), and norm is the normalization
   (*norm) of the path integral (which also includes the normalization of Gaussians).
   Array *norm is defined in main(), array *distrexp in distr().
   The improved potential Vp() for every pair of points (q[i], q[i + step])
   is evaluated at the mid-point qbar.
*/

double func(double *q, int s, int n)
{
   void V0(double, double); /* The user supplied function that returns the unimproved potential
                               and its derivatives. */
   double Vp(void); /* The function that, for a given p, returns the improved potential. */
   double tbar; /* Time mid-point. */
   double aaa;

   int i; /* Loop variable for the loop over the time steps. */
   int j; /* Loop variable for the loop defining deltas array. */

   action = 0; /* Initialiation of the action. */
   Nmax = two[s]; /* The maximal number of time steps. */
   Nc = two[n]; /* The current number of time steps. */
   step = two[s - n]; /* Index distance of path nodes at current bisection level. */
   epss[1] = T / Nc; /* Current time step and related variables. */
   epsinv = 1. / epss[1];
   tbar = ta - 0.5 * epss[1];
   for(i = 2; i < _p; i++) epss[i] = epss[1] * epss[i - 1];

   for(i = 0; i < Nmax; i += step) /* Loop over path nodes at current bisection level. */
   {
      tbar += epss[1];
      qbar = (q[i + step] + q[i]) * 0.5; /* Mid-point for nearest nodes. */
      deltas[1] = 0.5 * (q[i + step] - q[i]); /* Half-distance between nearest nodes. */
      deltas[2] = deltas[1] * deltas[1];
      for(j = 3; j < 2 * _p - 1; j++) deltas[j] = deltas[1] * deltas[j - 1];
      V0(qbar, tbar); /* Initialization of unimproved potential and its derivatives at qbar. */
      action += 2 * deltas[2] * epsinv + epss[1] * Vp() - distrexp[i]; /* Accumulation of the action minus
                                                                     the exponent of the Gaussian. */
   }

   return exp(- action + norm[n]); /* norm[n] includes the path integral norm for a
                                      given number of time steps 2^n and normalization of Gaussians
                                      used as probability density functions. */
}

/* For a chosen p, the function Vp() returns the value of the improved potential. */

#if _p == 1
double Vp(void)
{
   return -(ss[1]*xs[1]) + xs[2]/2;
}
#endif

#if _p == 2
double Vp(void)  
{
   return deltas[2]/6 + epss[1]/12 - ss[1]*xs[1] + xs[2]/2;
}
#endif

#if _p == 3
double Vp(void)  
{
   return deltas[2]/6 + epss[1]/12 - (cs[1]*deltas[1]*epss[1]*omegas[1])/6 - (epss[2]*ss[2])/24 - ss[1]*xs[1] + (epss[2]*ss[1]*xs[1])/12 + 
 (epss[2]*omegas[2]*ss[1]*xs[1])/24 + xs[2]/2 - (epss[2]*xs[2])/24;                        
}     
#endif

#if _p == 4
double Vp(void)  
{
   return deltas[2]/6 + epss[1]/12 - (deltas[2]*epss[2])/360 - epss[3]/360 - (cs[1]*deltas[1]*epss[1]*omegas[1])/6 - (epss[2]*ss[2])/24 - ss[1]*xs[1] + 
 (epss[2]*ss[1]*xs[1])/12 + (epss[2]*omegas[2]*ss[1]*xs[1])/24 + xs[2]/2 - (epss[2]*xs[2])/24;                        
}     
#endif

#if _p == 5
double Vp(void)  
{
   return deltas[2]/6 + epss[1]/12 - (deltas[2]*epss[2])/360 - epss[3]/360 - (cs[1]*deltas[1]*epss[1]*omegas[1])/6 + (cs[1]*deltas[1]*epss[3]*omegas[1])/720 - 
 (cs[2]*epss[4]*omegas[2])/1440 + (cs[1]*deltas[1]*epss[3]*omegas[3])/240 - (epss[2]*ss[2])/24 + (epss[4]*ss[2])/240 + 
 (epss[4]*omegas[2]*ss[2])/480 - ss[1]*xs[1] + (epss[2]*ss[1]*xs[1])/12 - (epss[4]*ss[1]*xs[1])/120 + (epss[2]*omegas[2]*ss[1]*xs[1])/24 - 
 (epss[4]*omegas[2]*ss[1]*xs[1])/480 - (epss[4]*omegas[4]*ss[1]*xs[1])/1920 + xs[2]/2 - (epss[2]*xs[2])/24 + (epss[4]*xs[2])/240;                        
}     
#endif

#if _p == 6
double Vp(void)  
{
   return deltas[2]/6 + epss[1]/12 - (deltas[2]*epss[2])/360 - epss[3]/360 + (deltas[2]*epss[4])/15120 + epss[5]/5670 - 
 (cs[1]*deltas[1]*epss[1]*omegas[1])/6 + (cs[1]*deltas[1]*epss[3]*omegas[1])/720 + (cs[1]*deltas[3]*epss[3]*omegas[1])/1008 + 
 (cs[1]*deltas[1]*epss[4]*omegas[1])/8064 - (cs[2]*epss[4]*omegas[2])/1440 + (cs[1]*deltas[1]*epss[3]*omegas[3])/240 - (epss[2]*ss[2])/24 + 
 (epss[4]*ss[2])/240 + (epss[4]*omegas[2]*ss[2])/480 - ss[1]*xs[1] + (epss[2]*ss[1]*xs[1])/12 - (epss[4]*ss[1]*xs[1])/120 + 
 (epss[2]*omegas[2]*ss[1]*xs[1])/24 - (epss[4]*omegas[2]*ss[1]*xs[1])/480 - (epss[4]*omegas[4]*ss[1]*xs[1])/1920 + xs[2]/2 - (epss[2]*xs[2])/24 + 
 (epss[4]*xs[2])/240;                        
}     
#endif

#if _p == 7
double Vp(void)  
{
   return deltas[2]/6 + epss[1]/12 - (deltas[2]*epss[2])/360 - epss[3]/360 + (deltas[2]*epss[4])/15120 + epss[5]/5670 - 
 (cs[1]*deltas[1]*epss[1]*omegas[1])/6 + (cs[1]*deltas[1]*epss[3]*omegas[1])/720 + (cs[1]*deltas[3]*epss[3]*omegas[1])/1008 + 
 (cs[1]*deltas[1]*epss[4]*omegas[1])/8064 - (cs[1]*deltas[1]*epss[5]*omegas[1])/40320 - (cs[2]*epss[4]*omegas[2])/1440 + 
 (cs[2]*epss[6]*omegas[2])/120960 + (cs[1]*deltas[1]*epss[3]*omegas[3])/240 - (cs[1]*deltas[1]*epss[5]*omegas[3])/40320 + 
 (cs[2]*epss[6]*omegas[4])/40320 - (cs[1]*deltas[1]*epss[5]*omegas[5])/26880 - (epss[2]*ss[2])/24 + (epss[4]*ss[2])/240 - (17*epss[6]*ss[2])/40320 + 
 (epss[4]*omegas[2]*ss[2])/480 - (epss[6]*omegas[2]*ss[2])/5040 - (epss[6]*omegas[4]*ss[2])/20160 - ss[1]*xs[1] + (epss[2]*ss[1]*xs[1])/12 - 
 (epss[4]*ss[1]*xs[1])/120 + (17*epss[6]*ss[1]*xs[1])/20160 + (epss[2]*omegas[2]*ss[1]*xs[1])/24 - (epss[4]*omegas[2]*ss[1]*xs[1])/480 + 
 (epss[6]*omegas[2]*ss[1]*xs[1])/5040 - (epss[4]*omegas[4]*ss[1]*xs[1])/1920 + (epss[6]*omegas[4]*ss[1]*xs[1])/53760 + 
 (epss[6]*omegas[6]*ss[1]*xs[1])/322560 + xs[2]/2 - (epss[2]*xs[2])/24 + (epss[4]*xs[2])/240 - (17*epss[6]*xs[2])/40320;                        
}     
#endif

#if _p == 8
double Vp(void)  
{
   return deltas[2]/6 + epss[1]/12 - (deltas[2]*epss[2])/360 - epss[3]/360 + (deltas[2]*epss[4])/15120 + epss[5]/5670 - (deltas[2]*epss[6])/604800 - 
 epss[7]/75600 - (cs[1]*deltas[1]*epss[1]*omegas[1])/6 + (cs[1]*deltas[1]*epss[3]*omegas[1])/720 + (cs[1]*deltas[3]*epss[3]*omegas[1])/1008 + 
 (cs[1]*deltas[1]*epss[4]*omegas[1])/8064 - (cs[1]*deltas[1]*epss[5]*omegas[1])/40320 - (cs[1]*deltas[3]*epss[5]*omegas[1])/30240 - 
 (29*cs[1]*deltas[1]*epss[6]*omegas[1])/7741440 - (cs[2]*epss[4]*omegas[2])/1440 + (cs[2]*epss[6]*omegas[2])/120960 + 
 (cs[2]*deltas[2]*epss[6]*omegas[2])/72576 + (5*cs[2]*epss[7]*omegas[2])/4644864 + (cs[1]*deltas[1]*epss[3]*omegas[3])/240 - 
 (cs[1]*deltas[1]*epss[5]*omegas[3])/40320 - (cs[1]*deltas[3]*epss[5]*omegas[3])/51840 - (cs[1]*deltas[1]*epss[6]*omegas[3])/552960 + 
 (cs[2]*epss[6]*omegas[4])/40320 - (cs[1]*deltas[1]*epss[5]*omegas[5])/26880 - (epss[2]*ss[2])/24 + (epss[4]*ss[2])/240 - (17*epss[6]*ss[2])/40320 + 
 (epss[4]*omegas[2]*ss[2])/480 - (epss[6]*omegas[2]*ss[2])/5040 - (epss[6]*omegas[4]*ss[2])/20160 - ss[1]*xs[1] + (epss[2]*ss[1]*xs[1])/12 - 
 (epss[4]*ss[1]*xs[1])/120 + (17*epss[6]*ss[1]*xs[1])/20160 + (epss[2]*omegas[2]*ss[1]*xs[1])/24 - (epss[4]*omegas[2]*ss[1]*xs[1])/480 + 
 (epss[6]*omegas[2]*ss[1]*xs[1])/5040 - (epss[4]*omegas[4]*ss[1]*xs[1])/1920 + (epss[6]*omegas[4]*ss[1]*xs[1])/53760 + 
 (epss[6]*omegas[6]*ss[1]*xs[1])/322560 + xs[2]/2 - (epss[2]*xs[2])/24 + (epss[4]*xs[2])/240 - (17*epss[6]*xs[2])/40320;                        
}     
#endif
